/*
    Spakes - a client/server Snakes-clone for up to 4 players
    Copyright (C) 2001, 2002, 2004  Markus Triska

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact me at triska@gmx.at
*/

#include <algorithm>
#include <cassert>

#include <qbuttongroup.h>
#include <qbrush.h>
#include <qkeycode.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qpalette.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qtimer.h>

#include "gameboard.h"
#include "player.h"
#include "server.h"
#include "event.h"
#include "protocol.h"
#include "udpsock.h"

using namespace std;

#include "mylog.h"
MyLogFile logfile;

extern Server server;
extern Player *players[4];
extern vector<Point> apples;
extern int numplayers;
// During STAT_WAITING, players may cancel.  "maxplayers" tells us where
// the last player in the array could reside, whereas "numplayers" holds
// the number of actual players.
int maxplayers = 0;
extern bool multiplayer;

const int STAT_LOGIN   = 0;              // First screen
const int STAT_WAITING = 1;              // Waiting for others to connect
const int STAT_PLAYING = 2;
const int STAT_PAUSE   = 3;
const int STAT_HISCORE = 4;

const int WINDOWW = 720;
const int WINDOWH = 450;


Gameboard::Gameboard(QWidget *parent, const char *name) : QWidget (parent, name)
{
	numplayers = 0;
	state = STAT_LOGIN;
	primplayerid = -1;
	secplayerid = -1;
	secplayer = PL_NONE;

	setCaption(QString("Spakes ") + VERSION);
	setPalette(QPalette(gray));
	setMinimumSize(WINDOWW, WINDOWH);
	setMaximumSize(WINDOWW, WINDOWH);

	txtPrimNick = new QLineEdit(this);
	txtPrimNick->setText("Nick");
	txtPrimNick->move(20, 20);
	txtPrimNick->show();

	txtIP = new QLineEdit(this);
	txtIP->setText("127.0.0.1");
	txtIP->move(20, 60);
	txtIP->show();

	btnSinglePlayer = new QPushButton("Local game", this);
	btnSinglePlayer->move(300, 15);
	btnSinglePlayer->show();
	connect(btnSinglePlayer, SIGNAL(clicked()), this, SLOT(start_singleplayer()));

	btnConnect = new QPushButton("Connect", this);
	btnConnect->move(300, 50);
	btnConnect->show();
	connect(btnConnect, SIGNAL(clicked()), this, SLOT(start_multiplayer()));

	secgroup = new QButtonGroup(3, Vertical, "Second player: ", this);
	secgroup->move(20, 110);
	secgroup->setMinimumSize(130, 120);
	secgroup->show();

	optAI = new QRadioButton("Computer", secgroup);
	optAI->show();
	optHuman = new QRadioButton("Human", secgroup);
	optHuman->show();
	optNone = new QRadioButton("None", secgroup);
	optNone->setChecked(true);
	optNone->show();

	udpserver = new UDPSocket();

	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(tick()));
}

Gameboard::~Gameboard()
{
	for (int i = 0; i < numplayers; i++) {
		if (players[i]) delete players[i];
	}
	if (multiplayer) server.logout();
	delete udpserver;
}

void Gameboard::cleanup_game()
{
	state = STAT_LOGIN;
	for (int i = 0; i < numplayers ; i++) {
		if (players[i]) {
			delete players[i];
			players[i] = 0;
		}
	}
	apples.clear();
	primplayerid = -1;
	secplayerid = -1;
	numplayers = 0;
	maxplayers = 0;
	show_controls();
	repaint();
}

void Gameboard::keyPressEvent(QKeyEvent *e)
{
	Event newevent;

	switch (state) {
	case STAT_PAUSE:
		state = STAT_PLAYING;
		break;
	case STAT_WAITING:
		if (e->key() == Key_Escape) {
			timer->stop();
			if (multiplayer) server.logout();
			cleanup_game();
			repaint();
		}
		break;

	case STAT_PLAYING:
		bool alldead;
		alldead = true;
		for (int i = 0; i < numplayers; i++) {
			if (players[i]->Living()) alldead = false;
		}

		if ((e->key() == Key_P) && (!multiplayer)) state = STAT_PAUSE;
		if ((e->key() == Key_Escape) || alldead) {
			timer->stop();
			if (multiplayer) server.logout();
			state = STAT_HISCORE;
			setPalette(QPalette(QColor(gray)));
			break;
		}

		newevent.value1 = -1;
		newevent.type = KEYDOWN;
		switch(e->key()) {
		case Key_Up:	newevent.value1 = UPARROW; break;
		case Key_Right: newevent.value1 = RIGHTARROW; break;
		case Key_Down:  newevent.value1 = DOWNARROW; break;
		case Key_Left:  newevent.value1 = LEFTARROW; break;
		default: e->ignore();
		}
		if (newevent.value1 > -1) players[primplayerid]->process_event(newevent);

		if (secplayerid < 0) break;

		newevent.value1 = -1;
		switch(e->key()) {
		case Key_Q: case Key_7: newevent.value1 = UPLEFTARROW; break;
		case Key_W: case Key_8: newevent.value1 = UPARROW; break;
		case Key_E: case Key_9: newevent.value1 = UPRIGHTARROW; break;
		case Key_D: case Key_6: newevent.value1 = RIGHTARROW; break;
		case Key_C: case Key_3: newevent.value1 = DOWNRIGHTARROW; break;
		case Key_X: case Key_2: newevent.value1 = DOWNARROW; break;
		case Key_Y: case Key_1: case Key_Z:
			newevent.value1 = DOWNLEFTARROW; break;
		case Key_A: case Key_4: newevent.value1 = LEFTARROW; break;
		}
		if (newevent.value1 > -1) players[secplayerid]->process_event(newevent);

		break;
	case STAT_HISCORE:
		if (e->key() == Key_Space) {
			cleanup_game();
		}
		break;
	}
}


void Gameboard::keyReleaseEvent(QKeyEvent *e)
{
	Event newevent;
	switch (state) {
	case STAT_PLAYING:
		newevent.value1 = -1;
		newevent.type = KEYUP;
		switch(e->key()) {
		case Key_Up:	newevent.value1 = UPARROW;    break;
		case Key_Right: newevent.value1 = RIGHTARROW; break;
		case Key_Down:  newevent.value1 = DOWNARROW;  break;
		case Key_Left:  newevent.value1 = LEFTARROW;  break;
		default: e->ignore();
		}
		if (newevent.value1 > -1) players[primplayerid]->process_event(newevent);

		break;
	}
}

void Gameboard::get_options()
{
	primplayername = txtPrimNick->text().latin1();
	if (primplayername == "") primplayername = "Coward";
	if (optAI->isChecked()) {
		secplayer = PL_AI;
	} else if (optHuman->isChecked()) {
		secplayer = PL_HUMAN;
	} else if (optNone->isChecked()) {
		secplayer = PL_NONE;
	}

	secplayername = "Computer";
	if (secplayer == PL_HUMAN) {
		// secplayername = txtSecNick->text();
		// if (secplayername == "") secplayername = "Coward";
		secplayername = "Player2";
	}
}

void Gameboard::start_singleplayer()
{
	get_options();
	// simulate a server message
	ostringstream ofakecmd1;
	ofakecmd1 << primplayername << " 0 25 3 1";
	istringstream fakecmd1(ofakecmd1.str());
	numplayers = 1;
	players[0] = new Consoleplayer(this, 0, fakecmd1);
	primplayerid = 0;

	ostringstream ofakecmd2;
	ofakecmd2 << secplayername << " 79 25 7";
	istringstream fakecmd2(ofakecmd2.str());
	switch(secplayer) {
	case PL_AI:
		players[1] = new AIPlayer(this, 1, fakecmd2);
		numplayers++; secplayerid = 1;
		break;
	case PL_HUMAN:
		players[1] = new Consoleplayer(this, 1, fakecmd2, false);
		numplayers++; secplayerid = 1; break;
	case PL_NONE: break;
	}
	hide_controls();

	srand(time(NULL));
	apples.clear();

	for (int i = 0; i < 10; i++) {
		Point applepos;
		do {
			applepos.x = rand() % 80;
			applepos.y = rand() % 50;
		} while (std::find(apples.begin(), apples.end(), applepos) != apples.end());
		apples.push_back(applepos);
	}

	multiplayer = false;

	setPalette(QPalette(QColor(0, 0, 0)));
	state = STAT_PLAYING;
	timer->start(100);
	repaint();
}

void Gameboard::hide_controls()
{
	btnConnect->hide();
	btnSinglePlayer->hide();
	txtIP->hide();
	txtPrimNick->hide();
	// txtSecNick->hide();
	secgroup->hide();
	this->setFocus();
}

void Gameboard::show_controls()
{
	btnConnect->show();
	btnSinglePlayer->show();
	txtIP->show();
	txtPrimNick->show();
	// txtSecNick->show();
	secgroup->show();
}

void Gameboard::start_multiplayer()
{
	get_options();
	if (!server.Connect(txtIP->text().latin1())) {
			QMessageBox::warning(this, "Spakes", "Could not connect to server", "Ok", "", 0, 0);
			return;
	}

	udpserver->Discard();
	server.login_primary(primplayername, udpserver->Port());
	multiplayer = true;
	hide_controls();
	state = STAT_WAITING;
	timer->start(100);
	repaint();
}

void Gameboard::tick()
{
	if (state == STAT_PAUSE) return;
	static bool in_timerproc = false;
	if (in_timerproc) return;
	in_timerproc = true;

	static string command;

	while (multiplayer && server.receive_command(command)) {
		istringstream scommand(command);
		int com_playerid, com_x, com_y, com_loginaccepted, serverversion;
		unsigned char com_comid;
		string com_playername;
		scommand >> com_comid;
		switch(com_comid) {
		case SRV_LOGINREPLY:
			scommand >> com_loginaccepted;
			if (!com_loginaccepted) {
				// Login rejected

			} else {
				scommand >> primplayerid;
				players[primplayerid] = new Consoleplayer(this,
primplayerid, scommand);
				numplayers++;
				if (numplayers > maxplayers) maxplayers = numplayers;
				if (secplayer != PL_NONE) server.login_secondary(secplayername);
				players[primplayerid]->announce();
			}
			break;

		case SRV_SECONDARY:
			scommand >> secplayerid;
			switch(secplayer) {
			case PL_AI: players[secplayerid] = new AIPlayer(this, secplayerid, scommand); break;
			case PL_HUMAN: players[secplayerid] = new Consoleplayer(this, secplayerid, scommand, false); break;
			case PL_NONE: assert(0);
			}
			numplayers++;
			if (numplayers > maxplayers) maxplayers = numplayers;
			players[secplayerid]->announce();
			break;

		case SRV_REMPRIMARY:
		case SRV_REMSECONDARY:
			scommand >> com_playerid;
			if (com_comid == SRV_REMPRIMARY) {
				players[com_playerid] = new RemotePrimaryPlayer(this, com_playerid, scommand);
			} else {
				players[com_playerid] = new Player(this, com_playerid, scommand);
			}
			numplayers++;
			if (numplayers > maxplayers) maxplayers = numplayers;
			players[com_playerid]->announce();
			break;

		case SRV_PLAYER_OUT:
			if (state == STAT_WAITING) {
				scommand >> com_playerid;
				delete players[com_playerid];
				players[com_playerid] = 0;
				numplayers--;
				repaint();
			}
			break;

		case SRV_GAMEINFO:
			scommand >> serverversion;
			if (serverversion > PROTOCOL_VERSION) {
				QMessageBox::warning(this, "Spakes", "This client is outdated. Please update to a new version.", "Ok", "", 0, 0);
				server.logout();
			} else if (serverversion < PROTOCOL_VERSION) {
				QMessageBox::warning(this, "Spakes", "Trying to connect to an outdated server. Please update to a new version.", "Ok", "", 0, 0);
				server.logout();
				state = STAT_LOGIN;
				show_controls();
			} else {
				server.send_version_ok();
			}
			break;

		case SRV_APPLE:
			scommand >> com_x >> com_y;
			{
				Point newapple(com_x, com_y);
				apples.push_back(newapple);
			}
			break;

		case SRV_GO:
			state = STAT_PLAYING;
			raise();
			setFocus();
			setPalette(QPalette(QColor(0, 0, 0)));
			break;
		}

		for (int i = 0; i < maxplayers; i++) {
			if (players[i]) players[i]->process_event(command);
		}
		command = "";
	}

 	string udpcommand;
	while(multiplayer && udpserver->Receive(udpcommand))
		for (int i= 0; i < numplayers; i++)
			players[i]->process_event(udpcommand);

	if ((primplayerid > -1) && (state == STAT_PLAYING)) {
		int localmax = players[primplayerid]->num_ticks();
		if (secplayerid > -1) {
			localmax = max(players[secplayerid]->num_ticks(), localmax);
		}

		int lowtic = localmax;
		for (int i = 0; i < numplayers; i++)
			if (players[i]->Living())
				lowtic = min(lowtic, players[i]->num_ticks());

		if ( localmax - lowtic <= 3) {
			players[primplayerid]->tick();
			server.send_tick(*players[primplayerid]);
			if (secplayerid > -1) {
				players[secplayerid]->tick();
				server.send_tick(*players[secplayerid]);
			}
			for (int i = 0; i < numplayers; i++) {
				players[i]->send_tick(players[primplayerid]);
				if (secplayerid > -1) players[i]->send_tick(players[secplayerid]);
			}
		}
	}


	in_timerproc = false;
}



void Gameboard::paintEvent(QPaintEvent *)
{
	switch (state) {
	case STAT_LOGIN:
		drawText(10, 300, "Copyright (C) 2001, 2002, 2004 Markus Triska     triska@gmx.at");
		drawText(10, 340, "Dedicated to S.P.              Happy Birthday!");
		break;

	case STAT_WAITING:
		for (int i = 0; i < maxplayers; i++) {
			if (players[i]) players[i]->announce();
		}
		break;

	case STAT_PLAYING:
		for (vector<Point>::iterator it = apples.begin(); it < apples.end(); it++) draw_apple(*it);
		for (int i = 0; i < numplayers; i++) {
			draw_player(players[i], true);
		}

		break;

	case STAT_HISCORE:

		drawText(10, 50, "Score");

		for (int i = 0; i < numplayers; i++) {
			ostringstream hiscore;
			hiscore << players[i]->Name() << ": " << players[i]->Score();
			drawText(10, 100 + i*15, hiscore.str().c_str());
		}

		drawText(10, 300, "Press SPACE to continue");

		break;

	}
}

void Gameboard::undraw_begin(const Player *p)
{
	Point lastspot = p->dots.front();

	QPainter paint(this);
	paint.setPen(black);
	paint.setBrush(QBrush(SolidPattern));
	paint.drawEllipse(lastspot.x*SPOTSIZE, lastspot.y*SPOTSIZE, SPOTSIZE, SPOTSIZE);

}

void Gameboard::draw_apple(const Point &p)
{
	QPainter paint(this);
	paint.setPen(QColor(0, 200, 200));
	QBrush applefill(QColor(0, 200, 200));
	paint.setBrush(applefill);
	paint.drawEllipse(p.x*SPOTSIZE, p.y*SPOTSIZE, SPOTSIZE, SPOTSIZE);
}



void Gameboard::draw_player(Player *p, bool fullredraw)
{
	if (!p->Living()) return;
	if (p->dots.size() < 1) return;

	QPainter paint(this);
	QColor c;

	switch(p->playerid) {
	case 0:	c.setRgb(0, 0, 255); break;
	case 1: c.setRgb(200, 0, 0); break;
	case 2: c.setRgb(0, 200, 0); break;
	case 3: c.setRgb(100, 100, 100); break;
	}

	QBrush ellfill(c);
	paint.setPen(c);
	paint.setBrush(ellfill);
	if (fullredraw) {
		for (list<Point>::iterator it = p->dots.begin(); it != p->dots.end(); it++) {
			Point spot = *it;
			paint.drawEllipse(spot.x*SPOTSIZE, spot.y*SPOTSIZE, SPOTSIZE, SPOTSIZE);
		}
	} else {
		Point topspot = p->dots.back();
		paint.drawEllipse(topspot.x*SPOTSIZE, topspot.y*SPOTSIZE, SPOTSIZE, SPOTSIZE);
	}
}

void Gameboard::undraw_player(Player *p)
{
	QPainter paint(this);
	paint.setPen(QColor(0, 0, 0));
	paint.setBrush(QBrush(SolidPattern));
	for (list<Point>::iterator ibegin = p->dots.begin(); ibegin != p->dots.end(); ibegin++) {
		paint.drawEllipse(ibegin->x*SPOTSIZE, ibegin->y*SPOTSIZE, SPOTSIZE, SPOTSIZE);
	}
}

void Gameboard::announce_player(const Player *p)
{

	Point textpos;

	switch(p->playerid) {
	case 0: textpos = Point(WINDOWW/3, WINDOWH/2); break;
	case 1: textpos = Point(2*WINDOWW/3, WINDOWH/2); break;
	case 2: textpos = Point(WINDOWW/2, WINDOWH/3); break;
	case 3: textpos = Point(WINDOWW/2, 2*WINDOWH/3); break;
	}

	drawText(textpos.x, textpos.y, p->name.c_str());
}

